"""
A module that contains the RoughGranulation class, which is used to provide the necessary
granulation operations required for rough set theory. This class is inherited by more specialized
classes, such as RoughApproximation, to provide the basic granulation operations.
"""

from typing import Union, Dict, Set
from collections.abc import Iterable
from collections import Counter

import graphviz
import igraph as ig


class RoughGranulation:
    """
    A class that represents the necessary granulation operations required for rough set theory. This
    class is inherited by more specialized classes, such as RoughApproximation, to provide the basic
    granulation operations.
    """

    def __init__(self):
        self.graph = ig.Graph(directed=True)
        self.index: int = 0
        # keys: hashed frozenset or attribute name (if given) mapped to attribute values
        self.attribute_table = {}

    def __getitem__(self, item: Union[str, int]) -> Dict[str, list]:
        vertex = self.graph.vs.find(type_eq=item)
        neighbor_vertices = self.graph.vs[self.graph.neighbors(vertex)]

        # get any vertices from vertex's neighbors that actively apply a relation upon 'vertex'
        relations = [
            vertex["type"] for vertex in neighbor_vertices if vertex["type"] is not None
        ]

        results = {}
        for relation in relations:
            if relation not in results:
                results[relation] = []
            results[relation].extend(
                [
                    category
                    for category in self / relation
                    if item in category and category not in results[relation]
                ]
            )
            if len(results[relation]) == 1:
                # remove the list if unnecessary; rough set expects no list
                results[relation] = results[relation][0]
        return results

    def __div_helper(self, other) -> frozenset:
        categories = []
        # the neighbors of this vertex are the equivalence classes
        equivalence_vertices = self.graph.vs.select(type_eq=other)
        for category in equivalence_vertices:
            nodes = self.graph.predecessors(category.index)
            vertices = self.graph.vs.select(nodes)
            elements = [vertex["type"] for vertex in vertices]
            categories.append(frozenset(elements))
        return frozenset(categories)

    def __truediv__(
        self, other: Union[str, Iterable]
    ) -> Union[Dict[str, Iterable], frozenset]:
        """
        Given a relation, obtain the equivalence classes.
        Args:
            other:

        Returns:
            A frozenset of equivalence classes, where each class is a frozenset.
        """
        if isinstance(other, Iterable) and not isinstance(
            other, str
        ):  # if is a list of relations
            results = {}
            for relation in other:
                results[relation] = self.__div_helper(relation)
            return results

        # else, when only given a single relation
        return self.__div_helper(other)

    @property
    def universe(self) -> frozenset:
        """
        Get the universe of discourse.

        Returns:
            A frozenset of elements in the universe of discourse.
        """
        return frozenset(self.graph.vs.select(layer_ne=None)["type"])

    def set_granules(
        self, nodes, source=None, is_input: Union[None, bool] = None
    ) -> None:
        """
        Adds the given nodes to the graph to a layer, and increments the saved indexed layer.

        Args:
            nodes: The nodes to be added.
            source: The source of the nodes.
            is_input: Whether this vertex should be considered 'input' (e.g., input granules or
            output granules). If None, then the 'input' attribute is not set. If True, then the
            'input' attribute is set to True. If False, then the 'input' attribute is set to False.
            Default is None.

        Returns:
            None
        """
        if nodes is not None and isinstance(nodes, Iterable):
            self.graph.add_vertices(
                len(nodes),
                attributes={
                    "type": list(nodes),
                    "layer": self.index,
                    "source": source,
                    "input": is_input,
                },
            )
            self.index += 1

    def create_compound_edges(self, args, target_vertices) -> list:
        """
        A helper method to self.add_parent_relation

        Args:
            args: A collection of lists, where each element in the collection (a list),
            stores the 'type' of the vertex in the graph.
            target_vertices: The vertices that have been created that represent
            the parent relationships.

        Returns:
            A list of edges to be added for the given relation (type) with respect to the given
            items (args).
        """
        edges = []
        for compound, target_vertex in zip(args, target_vertices):
            try:
                for node_id in compound:
                    self.create_compound_edges_helper(node_id, edges, target_vertex)
            except TypeError:  # the "compound" is already a source vertex
                self.create_compound_edges_helper(compound, edges, target_vertex)
        return edges

    def create_compound_edges_helper(self, source, edges, targets):
        """
        A helper method for the create_compound_edges method that creates edges between the
        'source' and 'targets', and appends them to the list 'edges'. If the given 'source' is not
        an igraph.Vertex, attempt to look up the vertex, and then iterate over the 'targets'. If a
        'target' at any moment is not an igraph.Vertex, attempt to look up the vertex.

        Args:
            source:
            edges:
            targets:

        Returns:
            A list of edges, where each edge is a 2-tuple in the form of (source, target)
        """
        if not isinstance(source, ig.Vertex):  # if the source is not a vertex
            try:
                source_vertex = self.graph.vs.find(
                    type_eq=source
                )  # try to find its vertex
            except ValueError as exception:  # no such vertex;
                raise ValueError(
                    f"A vertex could not be found in the graph: {source}."
                ) from exception
        else:
            source_vertex = source
        try:
            for target in targets:
                if not isinstance(target, ig.Vertex):  # if the source is not a vertex
                    target_vertex = self.graph.vs.find(
                        type_eq=target
                    )  # try to find its vertex
                else:
                    target_vertex = target

                edges.append((source_vertex.index, target_vertex.index))
        except TypeError:  # the "targets" is already a target vertex
            edges.append((source_vertex.index, targets.index))

    def add_parent_relation(self, attr_type, args) -> list:
        """
        Add a relation (attr_type) that references the provided items (args).

        Args:
            attr_type: The type of the relation, this can be a callable function as well
            (e.g., AlgebraicProduct).
            args: A collection of lists, where each element in the collection (a list),
            stores the 'type' of the vertex in the graph.

        Returns:
            A list of vertices that represent the parent relationships.
        """
        if isinstance(attr_type, str) and str.isdigit(attr_type):
            attr_type = int(attr_type)  # for saving/loading purposes

        vertices = []
        for _ in range(len(args)):
            vertices.append(self.graph.add_vertex(type=attr_type))

        edges = self.create_compound_edges(args, vertices)
        self.add_weighted_edges(edges)
        return vertices

    def add_weighted_edges(self, edges) -> None:
        """
        Add edges to the RoughGranulation.graph, with a weight that is equal to its frequency (in
        the argument, 'edges').

        Args:
            edges: A list of edges to be added.

        Returns:
            None
        """
        unique_edges_and_frequencies = Counter(edges)  # keep only the unique edges
        unique_edges, frequencies = (
            unique_edges_and_frequencies.keys(),
            unique_edges_and_frequencies.values(),
        )
        self.graph.add_edges(es=unique_edges, attributes={"weight": list(frequencies)})

    def export_visual(self, filename, file_format="png", engine="dot") -> None:
        """
        Creates and exports a visual of the graph using the format (file_format)
        and the layout (engine), as specified.

        Args:
            filename: The name of the file to save.
            file_format: The format of the file to save (e.g., 'png', 'svg').
            engine: The method to use when creating the layout of the graph (e.g., 'twopi', 'sfdp').

        Returns:
            None
        """
        self.graph.write(f=f"{filename}.dot")
        # for SelfOrganize
        # ig.plot(self.graph, target="llm.png",
        #         vertex_label=[repr(v["function"]) for v in self.graph.vs])
        file_formats = ["png", "svg", "svgz", "pdf"]
        if file_format in file_formats:  # https://graphviz.org/docs/outputs/
            render_engines = ["twopi", "sfdp", "dot"]
            if engine in render_engines:  # https://graphviz.org/docs/layouts/
                graphviz.render(
                    format=file_format, filepath=f"{filename}", engine=engine
                )
            else:
                raise UserWarning(
                    f"Exporting graph visual was unsuccessful. "
                    f"Please select a permitted format: {render_engines}."
                )
        else:
            raise UserWarning(
                f"Exporting graph visual was unsuccessful. "
                f"Please select a permitted format: {file_formats}."
            )

    def family_intersection(self, relative_to: set) -> frozenset:
        """
        Get the intersection of a family of sets/categories 'relative_to' some set of relations.

        Args:
            relative_to: A selection of relations.

        Returns:
            The family intersection.
        """

        categories = [
            next(iter(category)) for category in (self / relative_to).values()
        ]
        return frozenset.intersection(*categories)

    def family_union(self, relative_to: set) -> frozenset:
        """
        Get the union of a family of sets/categories 'relative_to' some set of relations.

        Args:
            relative_to: A selection of relations.

        Returns:
            The family union.
        """
        categories = [
            next(iter(category)) for category in (self / relative_to).values()
        ]
        return frozenset.union(*categories)

    def edges(self, relation: str) -> Set[frozenset]:
        """
        Get the 'name' (vertex attribute) of each vertex that interacts with the given relation.

        Args:
            relation: A relation.

        Returns:
            A set of frozensets, where each frozenset contains the 'name' of vertices that are
            neighbors which interact with .
        """
        # get the vertices that interact w/ relation
        vertices = self.graph.vs.select(type_eq=relation)

        return {
            frozenset(
                self.graph.vs[node]["type"] for node in self.graph.neighbors(rule_node)
            )
            for rule_node in vertices
        }



"""
Utility functions, such as for getting the powerset of an iterable.
"""

from collections.abc import Iterable
from itertools import chain, combinations


def powerset(iterable: Iterable, min_items: int):
    """
    Get the powerset of an iterable.

    Args:
        iterable: An iterable collection of elements.
        min_items: The minimum number of items that must be in each subset.

    Returns:
        The powerset of the given iterable.
    """
    # https://stackoverflow.com/questions/1482308/how-to-get-all-subsets-of-a-set-powerset
    return chain.from_iterable(
        combinations(list(iterable), r)
        for r in range(min_items, len(list(iterable)) + 1)
    )



"""
Implements the methods required to work with rough theory.
"""

from collections.abc import Iterable
from typing import Union, Tuple, List
from itertools import chain, combinations

from rough.approximation import RoughApproximation


def powerset(iterable: Iterable, min_items: int):
    """
    Get the powerset of an iterable.

    Args:
        iterable: An iterable collection of elements.
        min_items: The minimum number of items that must be in each subset.

    Returns:
        The powerset of the given iterable.
    """
    # https://stackoverflow.com/questions/1482308/how-to-get-all-subsets-of-a-set-powerset
    return chain.from_iterable(
        combinations(list(iterable), r)
        for r in range(min_items, len(list(iterable)) + 1)
    )


class RoughOperations(RoughApproximation):
    """
    This class implements several methods that are relevant to working with rough set theory.
    """

    # begin of granulation methods

    def y_dispensable(
        self, categories: Union[set, list], relation: str, category: str
    ) -> bool:
        """
        Determine if the category in categories is (relation)-dispensable in the intersection of
        categories.

        Args:
            categories: A set of categories.
            relation: A relation.
            category: A category.

        Returns:
            Whether the given category is dispensable.
        """
        if category in categories:
            (y_items,) = self.edges(relation)
            if self.family_intersection(categories).issubset(y_items):
                return self.family_intersection(categories - {category}).issubset(
                    y_items
                )
            raise ValueError(
                "The intersection of the family set 'categories' must be a subset of the given "
                "relation."
            )
        raise ValueError("The category must be an element of categories.")

    def y_independent(self, categories: set, category_y: str) -> bool:
        """
        The family of sets, categories, is Y-independent in the intersection of categories if all
        of its components are Y-indispensible in the intersection of categories; otherwise,
        the categories are Y-dependent in the intersection of categories.

        Args:
            categories: A set of categories.
            category_y:

        Returns:

        """
        for category in categories:
            if self.y_dispensable(categories, category_y, category):
                return False
        return True

    def is_y_reduct_of_f(self, categories_f, categories_h, category_y) -> bool:
        """
        The family H is a subset of F and a Y-reduct of the intersection of F,
        if H is Y-independent in the intersection of F, and the intersection of H
        is a subset of Y.

        Args:
            categories_f:
            categories_h:
            category_y:

        Returns:

        """
        if categories_h.issubset(categories_f):
            return self.y_independent(
                categories_h, category_y
            ) and self.family_intersection(categories_h).issubset(category_y)
        raise ValueError(
            "The set 'categories_h' must be a subset of the family set 'categories_f'."
        )

    # end of granulation methods

    # begin of metrics methods

    def dispensable(
        self,
        relations: Union[set, frozenset],
        relation: Union[str, set],
        relative_to: Union[set, frozenset] = None,
        mode: callable = None,
    ) -> bool:
        """
        Checks whether the relation is dispensable from the set "relations".

        Args:
            relations: The set of relations to check for dispensability.
            relation: The relation to check for dispensability.
            relative_to: The set of relations to find the relative positive region. If None, then
                the relative positive region is the positive region of the set "relations".
            mode: The mode to determine dispensability.

        Returns:
            True if dispensable, False otherwise.
        """
        if relative_to is None:
            try:
                return mode(relations, relative_to) == mode(
                    relations - frozenset(relation), relative_to
                )
            except TypeError:  # e.g., indiscernibility
                return mode(relations) == mode(relations - frozenset(relation))
        else:
            if relation in relations or (
                isinstance(relation, set)
                and len(relation) == 1
                and relation.intersection(relations)
            ):
                return self.find_relative_positive_region(
                    relations, relative_to
                ) == self.find_relative_positive_region(
                    relations - set(relation), relative_to
                )
            raise ValueError("The 'relation' must be an element of 'relations'.")

    def independent(
        self,
        relations: Union[set, frozenset],
        relative_to: Union[set, frozenset] = None,
        mode: callable = None,
    ) -> bool:
        """
        Checks each relation within the set "relations" to see if it is dispensable;
        if at least one relation is dispensable, then the set "relations" are dependent.
        Otherwise, the set of "relations" are independent.

        Args:
            relations: The set of relations to check for independence.
            relative_to: The set of relations to find the relative positive region. If None, then
            the relative positive region is the positive region of the set "relations".
            mode: The mode to determine dispensability. If None, then the mode is the
            indiscernibility.

        Returns:
            True if independent, False otherwise.
        """
        if mode is None:
            mode = self.indiscernibility
        for relation in relations:
            if self.dispensable(
                relations, relation, relative_to=relative_to, mode=mode
            ):
                return False
        return True

    # end of metrics methods

    # start of relations methods

    def __find_reducts_helper(
        self, relative_to, func, relative_target=None
    ) -> Tuple[set, list, set]:
        """
        A helper method for find_reducts().

        Args:
            relative_to:
            func:

        Returns:
            The indispensables, possible reducts, and target knowledge.
        """
        indispensables = set()
        dispensable_relations = set()

        for relation in relative_to:
            if self.dispensable(relative_to, relation, relative_target, func):
                dispensable_relations.add(relation)
            else:
                indispensables.add(relation)

        possible_reducts = list(powerset(relative_to, min_items=2))
        if relative_target is not None:
            possible_reducts = [
                frozenset(reduct)
                for reduct in possible_reducts
                if len(frozenset(reduct).intersection(indispensables)) > 0
            ]
        try:
            target_knowledge = func(relative_to, relative_target)
        except TypeError:  # e.g., indiscernibility
            target_knowledge = func(relative_to)
        return indispensables, possible_reducts, target_knowledge

    def find_reducts(
        self, relations: set, relative_to: set = None, mode: callable = None
    ) -> frozenset:
        """
        Find the reducts of the knowledge base given the "relations", relative to the relations
        specified in "relative_to", if applicable.

        Args:
            relations: The set of relations to find the reducts.
            relative_to: The set of relations to find the relative positive region. If None, then
            the relative positive region is the positive region of the set "relations".
            mode: The mode to determine dispensability. If relative_to is None and mode is None,
            then the mode is the indiscernibility.

        Returns:
            The set of reducts.
        """
        if relative_to is None:
            if mode is None:
                mode = self.indiscernibility
            (
                indispensables,
                possible_reducts,
                target_knowledge,
            ) = self.__find_reducts_helper(relations, mode, relative_to)
            # has to be independent and at least 1 indispensable relation
            return frozenset(
                [
                    frozenset(possible_reduct)
                    for possible_reduct in possible_reducts
                    if self.independent(frozenset(possible_reduct), mode=mode)
                    # and self.depends_on(relations, frozenset(possible_reduct))
                    and mode(possible_reduct) == target_knowledge
                    and len(indispensables.intersection(possible_reduct)) > 0
                ]
            )
        # calculate the relative REDUCT
        if mode is None:
            mode = self.find_relative_positive_region
        (
            indispensables,
            possible_reducts,
            target_knowledge,
        ) = self.__find_reducts_helper(relations, mode, relative_to)
        # has to be independent and at least 1 indispensable relation
        return frozenset(
            [
                frozenset(possible_reduct)
                for possible_reduct in possible_reducts
                if self.independent(frozenset(possible_reduct), relative_to=relative_to)
                # and self.depends_on(relative_to, frozenset(possible_reduct))
                and self.find_relative_positive_region(possible_reduct, relative_to)
                == target_knowledge
                and len(indispensables.intersection(possible_reduct)) > 0
            ]
        )

    def find_core(
        self, relations: set, relative_to: set = None, mode: callable = None
    ) -> frozenset:
        """
        Obtains the reducts of the knowledge and identifies the most important part aka the core.

        Args:
            relations:
            relative_to:
            mode: The function to use to calculate the core. Defaults to indiscernibility.

        Returns:

        """
        if relative_to is None or relations == relative_to:  # calculate CORE
            if mode is None:
                mode = self.indiscernibility
            return frozenset.intersection(
                *self.find_reducts(relations, relative_to=None, mode=mode)
            )
        # calculate the relative CORE (i.e., Q-CORE, where "Q" is relative_to)
        results = set()
        for relation in relations:
            if not self.dispensable(relations, relation, relative_to=relative_to):
                results.add(relation)
        return frozenset(results)

    def find_relative_positive_region(
        self, relations: Union[str, set], relative_to: Union[str, set]
    ) -> frozenset:
        """
        The P-positive region (POS) of Q is the set of all objects of the universe, U,
        which can be properly classified to classes of U / Q employing knowledge
        expressed by the classification U / P.

        Args:
            relations: The set of relations to find the relative positive region.
            relative_to:

        Returns:
            The relative positive region.
        """
        # categories = self / relative_to
        categories = self.indiscernibility(relative_to)
        results = set()
        for category in categories:
            results = results.union(self.lower_approximation(relations, category))
        return frozenset(results)

    def is_q_reduct_of_p(self, relations: set, restrictions: set, set_s: set) -> bool:
        """
        The family S that is a subset of P is called a Q-reduct of P,
        if and only if S is the Q-independent subfamily of P and
        POS_{S}(Q) == POS_{P}(Q).

        Args:
            relations:
            restrictions:
            set_s:

        Returns:

        """
        return self.find_relative_positive_region(
            set_s, restrictions
        ) == self.find_relative_positive_region(relations, restrictions)

    # end of relation methods

    # begin of dependency methods

    def depends_on(
        self, relations: Union[str, set], other_relations: Union[str, set]
    ) -> bool:
        """
        Knowledge Q depends on knowledge P (P ==> Q) if and
        only if IND(P) is a subset of IND(Q). More precisely,
        knowledge Q is derivable from knowledge P, if all
        elementary categories of Q can be defined in terms
        of some elementary categories of knowledge P.

        Args:
            relations:
            other_relations:

        Returns:

        """
        return all(
            # the elements of IND(P) must have at least 1 subset match in IND(Q)
            any(
                # the element in IND(P) is a subset of any element in IND(Q)
                P_element.issubset(Q_element)
                for Q_element in self.indiscernibility(other_relations)
            )
            for P_element in self.indiscernibility(relations)
        )

    def equivalent_to(
        self, relations: Union[str, set], other_relations: Union[str, set]
    ) -> bool:
        """
        Knowledge P and Q are equivalent if and only if
        knowledge P derives Q and knowledge Q derives P.

        Args:
            relations:
            other_relations:

        Returns:

        """
        return self.indiscernibility(relations) == self.indiscernibility(
            other_relations
        )

    def independent_of(
        self, relations: Union[str, set], other_relations: Union[str, set]
    ) -> bool:
        """
        Knowledge P and Q are independent, denoted as P != Q,
        if neither P ==> Q nor Q ==> P hold.

        Args:
            relations:
            other_relations:

        Returns:

        """
        # pylint: disable=arguments-out-of-order
        return not self.depends_on(relations, other_relations) and not self.depends_on(
            other_relations, relations
        )

    def partial_depends_on(
        self,
        relations: Union[str, set],
        other_relations: Union[str, set],
        category: Union[set, frozenset, List[frozenset]] = None,
    ) -> float:
        """
        Partial dependency of knowledge: the derivation (dependency)
        can also be partial, which means that only part of the knowledge Q
        is derivable from knowledge P.

        In essence, knowledge Q depends on a degree k (0 <= k <= 1) from
        knowledge P, symbolically P ==>_{k} Q.

        If k = 1, then Q totally depends on P; if 0 < k < 1, then Q roughly
        (partially) depends on P, and if k = 0, then Q is totally independent
        from knowledge P.

        In other words, if k = 1, then all the elements of the universe, U,
        can be classified to elementary categories of U / Q. If k != 1, only
        those elements of the universe which belong to the positive region
        can be classified to categories of knowledge Q, employing knowledge P.
        In particular, if k = 0, then none of the elements of the universe can
        be classified using knowledge P to elementary categories of knowledge Q.

        Disclaimer: The measure k of dependency, $P -->_{k} Q$, does not capture
        how this partial dependency is actually distributed among classes of
        U / Q. For example, some decision classes can be fully characterized by P,
        whereas others may be characterized only partially. Hence, an alternative
        definition is proposed where $gamma(X) = card(P_lower X) / card(X)$ where
        $X in U / Q$ which shows how many elements of each class of U / Q can be
        classified by employing knowledge P.

        To use the 'alternative definition', change the 'category' argument to a set
        that occurs in U / Q. When using the alternative definition, 'Q' can
        be 'None'.

        Args:
            relations:
            other_relations:
            category:

        Returns:
            The partial dependency of knowledge.
        """
        if category is not None:
            # pylint: disable=fixme
            # TODO: check this is reached by code coverage
            return len(self.lower_approximation(relations, category)) / len(category)
        return len(
            self.find_relative_positive_region(relations, other_relations)
        ) / len(self.universe)

    # end of dependency methods



"""
A module that contains the RoughApproximation class, which is a class that represents the rough
approximation of a set of objects. The rough approximation is a set of objects that are
indiscernible from each other with respect to a set of relations. The rough approximation is a
generalization of the lower and upper approximation, which are defined by Pawlak on page 10 of his
book "Rough Sets: Theoretical Aspects of Reasoning About Data".
"""

from collections import namedtuple
from collections.abc import Iterable
from typing import List, Union, Set

from rough.granulation import RoughGranulation


class RoughApproximation(RoughGranulation):
    """
    A class that represents the rough approximation of a set of objects. The rough approximation is
    a set of objects that are indiscernible from each other with respect to a set of relations. The
    rough approximation is a generalization of the lower and upper approximation, which are defined
    by Pawlak on page 10 of his book "Rough Sets: Theoretical Aspects of Reasoning About Data".
    """

    def indiscernibility(self, equivalence_relations: Union[str, set, list]):
        """
        The indiscernibility relation over a vector of equivalence relations (referred to as
        "P"). Defined on page 3 of "Rough Sets: Theoretical Aspects of Reasoning About Data".

        Args:
            equivalence_relations: An iterable of equivalence relations that must be a subset
            of the available relations and cannot be empty. If a string is passed, it is assumed to
            be a single equivalence relation. If a set is passed, it is assumed to be a set of
            equivalence relations. If a list is passed, it is assumed to be a list of equivalence
            relations.

        Returns:
            An indiscernibility relation over a collection of equivalence relations (i.e.,
            the family of equivalence classes of the equivalence relation IND(P)) yields the
            family of equivalence classes; denotes knowledge associated with the family of
            equivalence relations P, called P-basic knowledge about U (universe) in K (knowledge
            base).
        """
        if len(equivalence_relations) == 0:
            raise ValueError("The relations' length must be greater than zero.")

        possible_relations = self.graph.vs.select(type_ne=None)
        relation_names = frozenset(
            [relation["type"] for relation in possible_relations]
        )
        if frozenset(equivalence_relations).issubset(relation_names):
            indiscernibility_relation = set()
            for element in self.universe:
                categories = self[element]
                if len(categories) > 0:
                    # it is possible for self[element] to have no equivalence relations
                    new_category = set(self.universe)
                    for relation in equivalence_relations:
                        if relation in categories:
                            # some elements might not be defined for all relations
                            new_category = new_category.intersection(
                                categories[relation]
                            )
                    indiscernibility_relation.add(frozenset(new_category))
            return indiscernibility_relation
        raise ValueError(
            "The relations must be a subset of the existing relations on the graph."
        )

    def lower_approximation(
        self,
        relations: Union[str, set],
        categories: Union[set, frozenset, List[frozenset]],
    ) -> Union[frozenset, Set[frozenset]]:
        """
        Get the lower approximation of a (set of) category(s), as defined by Pawlak on page 10 of
        his book "Rough Sets: Theoretical Aspects of Reasoning About Data".

        Args:
            relations: Either a set of relations or a string that references a specific relation
            in the RoughApproximation.
            categories: The category (frozenset) or a family of categories (list of frozensets).

        Returns:
            Either a frozenset or a set of frozensets that represent the lower approximation.
        """

        return self.approximation(
            relations,
            categories,
            mode=lambda subset, category: subset.issubset(category),
        )

    def upper_approximation(
        self,
        relations: Union[str, set],
        categories: Union[set, frozenset, List[frozenset]],
    ) -> Union[frozenset, Set[frozenset]]:
        """
        Get the upper approximation of a (set of) category(s), as defined by Pawlak on page 10 of
        his book "Rough Sets: Theoretical Aspects of Reasoning About Data".

        Args:
            relations: Either a set of relations or a string that references a specific relation
            in the RoughApproximation.
            categories: The category (frozenset) or a family of categories (list of frozensets).

        Returns:
            Either a frozenset or a set of frozensets that represent the upper approximation.
        """
        return self.approximation(
            relations,
            categories,
            mode=lambda subset, category: len(subset.intersection(category)) > 0,
        )

    def approximation(
        self,
        relations: Union[str, set],
        categories: Union[set, frozenset, List[frozenset]],
        mode: callable,
    ) -> Union[frozenset, Set[frozenset]]:
        """
        Get the approximation of a (set of) category(s), as defined by Pawlak on page 10 of his
        book "Rough Sets: Theoretical Aspects of Reasoning About Data". The approximation is
        determined by the mode argument, which is a function that takes two arguments: a subset
        and the category. The mode function should return True if the subset satisfies the category.

        Args:
            relations: Either a set of relations or a string that references a specific relation
            in the RoughApproximation.
            categories: The category (frozenset) or a family of categories (list of frozensets).
            mode: The mode of approximation, which is a function that takes two arguments: a
            subset and a category. The mode function should return True if the subset satisfies
            the category. In the case of the lower approximation, the mode function should return
            True if the subset is a subset of the category. In the case of the upper
            approximation, the mode function should return True if the subset intersects with the
            category.

        Returns:
            Either a frozenset or a set of frozensets that represent the [mode] approximation.
        """

        def __approximation(
            relations: Union[str, set],
            category: frozenset,
            satisfied_constraint: callable,
        ) -> frozenset:
            """
            Get the approximation of a category.

            Args:
                relations: Either a set of relations or a string that references a specific relation
                in the RoughApproximation.
                categories: The category (frozenset) or a family of categories (list of frozensets).
                satisfied_constraint: The constraint that is to be satisfied or otherwise
                known as the mode of approximation, which is a function that takes two arguments: a
                subset and a category. The mode function should return True if the subset satisfies
                the category. In the case of the lower approximation, the mode function should
                return True if the subset is a subset of the category. In the case of the upper
                approximation, the mode function should return True if the subset intersects with
                the category.

            Returns:
                A frozenset of the approximation.
            """
            indiscernibility_relation = self.indiscernibility(relations)
            result = set()
            for subset in indiscernibility_relation:
                if satisfied_constraint(subset, category):
                    result = result.union(subset)
            return frozenset(result)

        if len(categories) == 0:
            raise ValueError("The argument 'categories' may not have a length of zero.")
        if isinstance(categories, Iterable) and isinstance(categories, list):
            # categories is a family of non-empty sets
            result = set()
            for category in categories:
                if len(category) == 0:
                    raise ValueError(
                        "The argument 'categories' may not have an element with a length of zero."
                    )
                result.add(__approximation(relations, category, mode))
        elif isinstance(categories, (set, frozenset)):
            result = __approximation(relations, categories, mode)
        else:
            raise ValueError(
                "The argument 'categories' must be a set, a frozenset, or a list."
            )

        return frozenset(result)

    def positive_region(
        self,
        relations: Union[str, set],
        categories: Union[set, frozenset, List[frozenset]],
    ) -> Union[frozenset, Set[frozenset]]:
        """
        Analogous to the R-lower approximation.

        Args:
            relations: Either a set of relations or a string that references a specific relation
            in the RoughApproximation.
            categories: The category (frozenset) or a family of categories (list of frozensets).

        Returns:
            Either a frozenset or a set of frozensets that represent the positive region or
            the lower approximation.
        """
        return self.lower_approximation(relations, categories)

    def negative_region(
        self, relations: Union[str, set], categories: Union[frozenset, List[frozenset]]
    ) -> Union[frozenset, Set[frozenset]]:
        """
        The set of objects with which it can be determined without
        any ambiguity, employing knowledge 'relations', that they
        do not belong to the 'categories' (i.e., they belong to the
        complement of 'categories')

        Args:
            relations: Either a set of relations or a string that references a specific relation
            in the RoughApproximation.
            categories: The category (frozenset) or a family of categories (list of frozensets).

        Returns:
            Either a frozenset or a set of frozensets that represent the negative region.
        """
        return self.universe - self.upper_approximation(relations, categories)

    def boundary_region(
        self,
        relations: Union[str, set],
        categories: Union[set, frozenset, List[frozenset]],
    ) -> Union[frozenset, Set[frozenset]]:
        """
        The undecidable area of the universe (i.e., none of the objects
        belonging to the boundary can be classified with certainty into
        categories or not categories, as far as knowledge 'relations' is concerned.

        Args:
            relations: Either a set of relations or a string that references a specific relation
            in the RoughApproximation.
            categories: The category (frozenset) or a family of categories (list of frozensets).

        Returns:
            Either a frozenset or a set of frozensets that represent the boundary region.
        """
        return self.upper_approximation(
            relations, categories
        ) - self.lower_approximation(relations, categories)

    def definable(self, relations: Union[str, set], category: frozenset) -> namedtuple:
        """
        A category X is called:

        R-definable if X is the union of some R-basic categories;
        otherwise X is R-undefinable. Also called R-exact sets. R-undefinable sets are also
        called R-inexact or R-rough sets.

        Roughly definable if we are able to decide whether some elements of the
        universe belong to X or not X.

        Internally undefinable if we are able to decide whether some elements of the
        universe belong to not X, but we are unable to indicate one element of X.

        Externally undefinable if we are able to decide for some elements of the
        universe whether they belong to X, but we are unable to indicate one element of not X.

        Totally undefinable if we are unable to decide for any element of the
        universe whether it belongs to X or not X.

        Args:
            relations: Either a set of relations or a string that references a specific relation
            in the RoughApproximation.
            category: The category (frozenset).

        Returns:
            A namedtuple with the lower and upper approximation of the given category. The name
            of the tuple depends on the type of definability of the category.
        """
        lower_approximation = self.lower_approximation(relations, category)
        upper_approximation = self.upper_approximation(relations, category)

        if lower_approximation == upper_approximation:
            # The set X is called R-definable if X is the union of some R-basic categories;
            # otherwise X is R-undefinable. Also called R-exact sets. R-undefinable sets are also
            # called R-inexact or R-rough sets.
            return namedtuple(
                "Definable", ["lower_approximation", "upper_approximation"]
            )(lower_approximation, upper_approximation)
        if (
            len(self.lower_approximation(relations, category)) > 0
            and self.upper_approximation(relations, category) != self.universe
        ):
            # We are able to decide whether some elements of the universe belong to X or not X.
            return namedtuple(
                "RoughlyDefinable", ["lower_approximation", "upper_approximation"]
            )(lower_approximation, upper_approximation)
        if (
            len(self.lower_approximation(relations, category)) == 0
            and self.upper_approximation(relations, category) != self.universe
        ):
            # We are able to decide whether some elements of the universe belong to not X,
            # but we are unable to indicate one element of X.
            return namedtuple(
                "InternallyUndefinable", ["lower_approximation", "upper_approximation"]
            )(lower_approximation, upper_approximation)
        if (
            len(self.lower_approximation(relations, category)) != 0
            and self.upper_approximation(relations, category) == self.universe
        ):
            # We are able to decide for some elements of the universe whether they belong to X,
            # but we are unable to indicate one element of not X.
            return namedtuple(
                "ExternallyUndefinable", ["lower_approximation", "upper_approximation"]
            )(lower_approximation, upper_approximation)
        if (
            len(self.lower_approximation(relations, category)) == 0
            and self.upper_approximation(relations, category) == self.universe
        ):
            # We are unable to decide for any element of the universe whether
            # it belongs to X or not X.
            return namedtuple(
                "TotallyUndefinable", ["lower_approximation", "upper_approximation"]
            )(lower_approximation, upper_approximation)
        raise ValueError("The given category is not definable or roughly definable.")

    def quality_of_approximation(self, relations, categories):
        """
        The quality of approximation of X, a family of non-empty sets,
        using knowledge 'relations'.

        This metric expresses the percentage of possible correct
        decisions when classifying objects employing the knowledge 'relations'.

        Args:
            relations:
            categories:

        Returns:

        """
        if len(categories) == 0:
            raise ValueError("The argument 'categories' may not have a length of zero.")
        if isinstance(categories, Iterable) and isinstance(categories, list):
            # categories is a family of non-empty sets
            numerator = 0.0
            for category in categories:
                if len(category) == 0:
                    raise ValueError(
                        "The argument 'category' may not have an element with a length of zero."
                    )
                numerator += len(self.lower_approximation(relations, category))
        else:
            numerator = len(self.lower_approximation(relations, categories))
        denominator = len(self.universe)

        return numerator / denominator

    def roughly_equal(
        self,
        relations,
        category,
        other_category,
        mode: Union["bottom", "top", "both"] = "both",
    ):
        """
        Determines if the category and the other category are roughly equal given knowledge about
        "relations". The definition of equality depends on the mode.

        If mode is "bottom", then the lower approximation of category is equal to the lower
        approximation of the other category given "relations"; essentially, the positive examples of
        the sets "category" and the "other category" are the same.

        If mode is "top", then the upper approximation of the given category is equal to the
        upper approximation of the other category given knowledge about "relations"; essentially,
        the negative examples of the sets "category" and the "other category" are the same.

        If mode is "both", then the lower and upper approximation of category is equal to the lower
        and upper approximation of the other category given "relations"; essentially, the positive
        and negative examples of the sets "category" and the "other category" are the same.

        Args:
            relations: The equivalence relations.
            category: The category in question.
            other_category: The other category in question.
            mode: The mode of equality.

        Returns:
            Whether the category and the other category are roughly equal given knowledge about
            "relations".
        """
        if mode == "bottom":
            return self.lower_approximation(
                relations, category
            ) == self.lower_approximation(relations, other_category)
        if mode == "top":
            return self.upper_approximation(
                relations, category
            ) == self.upper_approximation(relations, other_category)
        if mode == "both":
            return self.roughly_equal(
                relations, category, other_category, mode="bottom"
            ) and self.roughly_equal(relations, category, other_category, mode="top")

        raise ValueError("The argument 'mode' is not valid.")

    def roughly_included(
        self,
        relations,
        category,
        other_category,
        mode: Union["bottom", "top", "both"] = "both",
    ):
        """
        Determines whether the category is roughly included in the other category given knowledge
        about "relations". The definition of inclusion depends on the mode.

        If mode is "bottom", then the lower approximation of category is equal to the lower
        approximation of the other category given "relations"; essentially, the positive examples of
        the sets "category" and the "other category" are the same.

        If mode is "top", then the upper approximation of the given category is equal to the
        upper approximation of the other category given knowledge about "relations"; essentially,
        the negative examples of the sets "category" and the "other category" are the same.

        If mode is "both", then the lower and upper approximation of category is equal to the lower
        and upper approximation of the other category given "relations"; essentially, the positive
        and negative examples of the sets "category" and the "other category" are the same.

        Whether the lower and upper approximations of X are subset to the
        lower and upper approximations of Y given 'relations', respectively.

        Args:
            relations: The equivalence relations.
            category: The category in question.
            other_category: The other category in question.
            mode: The mode of inclusion (i.e., subsethood).

        Returns:
            Whether the category is roughly included in (i.e., a subset of) the other category
            given knowledge about "relations".
        """
        if mode == "bottom":
            return self.lower_approximation(relations, category).issubset(
                self.lower_approximation(relations, other_category)
            )
        if mode == "top":
            return self.upper_approximation(relations, category).issubset(
                self.upper_approximation(relations, other_category)
            )
        if mode == "both":
            return self.roughly_included(
                relations, category, other_category, mode="bottom"
            ) and self.roughly_included(relations, category, other_category, mode="top")

        raise ValueError("The argument 'mode' is not valid.")

    def accuracy(self, relations, category):
        """
        Calculate the inexactness of a set (category) due to the
        existence of a borderline region. The greater the borderline region,
        the lower the accuracy is of the set.

        If the accuracy is 1, then the set X is R-definable, otherwise,
        it is R-undefinable (i.e., our knowledge is incomplete).

        Args:
            relations:
            category:

        Returns:

        """
        if len(category) == 0:
            raise ValueError("The argument 'category' may not have a length of zero.")
        if isinstance(category, Iterable) and isinstance(category, list):
            # category is a family of non-empty sets
            numerator, denominator = 0.0, 0.0
            for set_x_i in category:
                if len(set_x_i) == 0:
                    raise ValueError(
                        "The argument 'category' may not have an element with a length of zero."
                    )
                numerator += len(self.lower_approximation(relations, set_x_i))
                denominator += len(self.upper_approximation(relations, set_x_i))
        else:
            numerator = len(self.lower_approximation(relations, category))
            denominator = len(self.upper_approximation(relations, category))
        return numerator / denominator

    def roughness(self, relations, category):
        """
        The complement of a rough set's accuracy or inexactness; represents
        the degree of incompleteness of knowledge 'relations' about the set X.

        Args:
            relations:
            category:

        Returns:

        """
        return 1 - self.accuracy(relations, category)



"""
Implements the methods required to work with rough theory.
"""

import itertools

from rough.operations import RoughOperations


class RoughDecisions(RoughOperations):
    """
    This class implements the necessary methods for handling decisions in rough set theory. For
    example, the discernibility matrix, the CORE, and the reduct. Features include the ability to
    simplify decision tables as well as remove redundant attributes.
    """

    # begin of discernibility methods

    def __delete_indiscernable_comparison(self, matrix, key, relations):
        if (
            any(
                key.issubset(decision_class)
                for decision_class in self.indiscernibility(relations)
            )
            and key in matrix
        ):
            del matrix[key]

    def __discernibility_matrix_helper(self, matrix, relations, func: callable) -> dict:
        for relation in relations:
            equivalence_classes = self / relation
            for equivalence_class in equivalence_classes:
                pairs = itertools.combinations(equivalence_class, r=2)
                for item_1, item_2 in pairs:
                    key = frozenset({item_1, item_2})
                    func(matrix, key, relation)
        return matrix

    def discernibility_matrix(self, relations, decision_attributes=None) -> dict:
        """
        Calculate the discernibility matrix for a set of relations.

        Args:
            relations: The set of relations.
            decision_attributes: The decision attributes.

        Returns:
            The discernibility matrix.
        """

        def delete_dispensable_relation(matrix, key, relations):
            if key in matrix:
                matrix[key] -= frozenset(relations)

        keys = itertools.combinations(self.universe, r=2)
        matrix = {
            frozenset(key): (
                frozenset(relations)
                if len(set(key)) > 1 and key[0] != key[1]
                else set()
            )
            for key in keys
        }

        if decision_attributes is not None:
            # Remove any pair-wise comparisons between objects that appear indiscernable
            # w.r.t. the decision attributes.
            self.__discernibility_matrix_helper(
                matrix, decision_attributes, func=self.__delete_indiscernable_comparison
            )

        return self.__discernibility_matrix_helper(
            matrix, relations, func=delete_dispensable_relation
        )

    def find_core_by_matrix(self, matrix) -> frozenset:
        """
        Find the CORE by the matrix method. The CORE is the set of attributes that are not
        redundant.

        Args:
            matrix: The discernibility matrix.

        Returns:
            The CORE.
        """
        relation_subsets = list(matrix.values())  # all subsets found in the matrix
        relation_subsets.sort(
            key=len
        )  # smaller sets start at 0, larger sets near end of list
        # each matrix entry that is a singleton set belongs to the CORE set
        return frozenset.union(
            *[relations for relations in relation_subsets if len(relations) == 1]
        )

    def __find_reduct_by_matrix(self, matrix):
        relation_subsets = list(matrix.values())  # all subsets found in the matrix
        relation_subsets.sort(
            key=len
        )  # smaller sets start at 0, larger sets near end of list

        core = self.find_core_by_matrix(matrix)

        # any matrix entry that has cardinality > 1 can possibly be used as a reduct
        possible_reducts = [
            relations for relations in relation_subsets if len(relations) > 1
        ]

        # now apply the CORE, wherever it is possible to choose CORE to
        # discern an object from another, do so

        for key, values in matrix.items():
            new_value = values.intersection(core)
            if len(new_value) > 0:  # check if the CORE may be used here
                matrix[key] = new_value  # replace the value with a CORE attribute
            else:  # find the next smallest subset and
                # update the matrix pair-wise comparison with it
                for possible_reduct in possible_reducts:
                    if possible_reduct.issubset(matrix[key]):
                        matrix[key] = matrix[key].intersection(possible_reduct)

        return matrix

    def minimum_discernibility_matrix(
        self, relations, matrix=None, decision_attributes=None
    ):
        """
        A. Skowron has proposed to represent knowledge in this form. It is a n x n matrix,
        where 'n' is the number of items in the universe of discourse. Each entry, c_{ij},
        in the matrix follows:

        $(c_{ij}) = { a in A : a(x_{i}) != a(x_{j}) } for i, j = 1, 2, ..., n$

        Thus entry c_{ij} is the set of all attributes which discern objects x_{i} and x_{j}.

        Args:
            relations:
            matrix:
            decision_attributes:

        Returns:

        """
        if matrix is None:  # if the matrix has not been calculated yet
            matrix = self.discernibility_matrix(
                relations, decision_attributes
            )  # then calculate it

        return self.__find_reduct_by_matrix(matrix)

    # end of discernibility methods

    # begin of decision table methods

    def decompose_decision_table(self, condition_attributes, decision_attributes):
        """
        Given the condition attributes (condition_attributes) and the decision attributes
        (decision_attributes), decompose the decision table into a consistent decision
        table and an inconsistent decision table.

        Args:
            condition_attributes: The condition attributes.
            decision_attributes: The decision attributes.

        Returns:
            A consistent decision table (frozenset), an inconsistent decision table (frozenset)
        """
        consistent_table = self.find_relative_positive_region(
            condition_attributes, decision_attributes
        )
        inconsistent_table = frozenset.union(
            *[
                self.boundary_region(condition_attributes, X)
                for X in self.indiscernibility(decision_attributes)
            ]
        )
        return consistent_table, inconsistent_table

    def find_attribute_cores(self, minimal_condition_attributes):
        """
        Find the core attributes.

        Args:
            minimal_condition_attributes:

        Returns:

        """
        core_attributes = {}

        for rule_idx in self.universe:
            attr_partitions = self[rule_idx]
            num_of_condition_attributes = len(minimal_condition_attributes) - 1
            condition_attributes_combinations = itertools.combinations(
                minimal_condition_attributes, num_of_condition_attributes
            )

            for selected_condition_attributes in condition_attributes_combinations:
                selected_condition_attributes = frozenset(selected_condition_attributes)
                family_of_condition_attributes = {
                    key: value
                    for key, value in attr_partitions.items()
                    if key in selected_condition_attributes
                }
                decision_category = frozenset.intersection(
                    *family_of_condition_attributes.values()
                )
                if not decision_category.issubset(attr_partitions["e"]):
                    # by only using the selected_condition_attributes,
                    # the decision_category has changed;
                    # this means that whatever we got rid of was actually important
                    if rule_idx not in core_attributes:
                        core_attributes[rule_idx] = set()
                    missing_category = (
                        minimal_condition_attributes - selected_condition_attributes
                    )
                    core_attributes[rule_idx] = core_attributes[rule_idx].union(
                        missing_category
                    )
        return core_attributes

    def find_attribute_reducts(self, minimal_condition_attributes):
        """
        Find the reduct attributes.

        Args:
            minimal_condition_attributes:

        Returns:

        """
        reduct_attributes = {}

        for rule_idx in self.universe:
            attr_partitions = self[rule_idx]

            for num_of_condition_attributes in range(
                1, len(minimal_condition_attributes)
            ):
                condition_attributes_combinations = itertools.combinations(
                    minimal_condition_attributes, num_of_condition_attributes
                )

                for selected_condition_attributes in condition_attributes_combinations:
                    selected_condition_attributes = frozenset(
                        selected_condition_attributes
                    )
                    family_of_condition_attributes = {
                        key: value
                        for key, value in attr_partitions.items()
                        if key in selected_condition_attributes
                    }
                    decision_category = frozenset.intersection(
                        *family_of_condition_attributes.values()
                    )

                    if decision_category.issubset(attr_partitions["e"]):
                        if rule_idx not in reduct_attributes:
                            reduct_attributes[rule_idx] = set()
                        # elif rule_idx not in reduct_done or not reduct_done[rule_idx]:
                        reduct_attributes[rule_idx].add(selected_condition_attributes)

                if rule_idx in reduct_attributes:  # we only want the smallest reducts
                    break
        return reduct_attributes

    def remove_redundant_attributes(self, condition_attributes, decision_attributes):
        """
        Remove any attributes identified as redundant.

        Args:
            condition_attributes: The condition attributes.
            decision_attributes: The decision attributes.

        Returns:

        """
        (result,) = self.find_reducts(
            condition_attributes, relative_to=decision_attributes
        )
        return result

    def simplify_decision_table(self, condition_attributes, decision_attributes):
        """
        Simplify the decision table, given the condition attributes and the decision attributes.

        Args:
            condition_attributes: The condition attributes.
            decision_attributes: The decision attributes.

        Returns:

        """
        minimal_condition_attributes = self.remove_redundant_attributes(
            condition_attributes, decision_attributes
        )
        return self.find_attribute_cores(
            minimal_condition_attributes
        ), self.find_attribute_reducts(minimal_condition_attributes)

    # end of decision table methods
